// ======================================================================
//
// MenuStaticListBox.cpp
// hugh david & ~Chuck Karpiak
//
// copyright 1997 origin systems inc
//
// ======================================================================

#include "First.h"
#include "MenuStaticListBox.h"

#include "Iff.h"
#include "MenuText.h"
#include "Mouse.h"
#include "VString.h"

#include <math.h>
#include <stdlib.h>

// ======================================================================

const Tag TAG_TXEN = TAG(T,X,E,N);
const Tag TAG_LBF  = TAG3(L,B,F);
const Tag TAG_LBFS = TAG(L,B,F,S);
const Tag TAG_LBEV = TAG(L,B,E,V);
const Tag TAG_LBSE = TAG(L,B,S,E);
const Tag TAG_LBTH = TAG(L,B,T,H);
const Tag TAG_LBHF = TAG(L,B,H,F);
const Tag TAG_LBHS = TAG(L,B,H,S);
const Tag TAG_JSTY = TAG(J,S,T,Y);

// ======================================================================

MenuStaticListBox::Link::Link(void)
{
	string         = NULL;
	overridingFont = NULL;
	prev           = NULL;
	next           = NULL;
}

// ----------------------------------------------------------------------

MenuStaticListBox::Link::~Link(void)
{
	delete string;
	string         = NULL;
	overridingFont = NULL;
	prev           = NULL;
	next           = NULL;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::Link::setString(const String &str, Font *overrideFont)
{
	*string = str;
	overridingFont = overrideFont;
}

// ======================================================================

MenuStaticListBox::MenuStaticListBox(MenuElement *newParent, int maxWidth, int maxHeight, Iff *iff)
: MenuElement(newParent, maxWidth, maxHeight, NULL)
{
	init();
	load(iff);
}

// ----------------------------------------------------------------------

MenuStaticListBox::~MenuStaticListBox(void)
{
	Link *link;

	for (link = firstLink; link; link = firstLink)
	{
		firstLink = link->next;
		delete link;
		link = NULL;
	}

	delete [] stringSelectEvent;
	stringSelectEvent = NULL;

	init();
}

// ----------------------------------------------------------------------

void MenuStaticListBox::init(void)
{
	visibleLinks      = 0;
	stringSelectEvent = NULL;
	numStringEvents   = NULL;
	font              = NULL;
	fontHighlight     = NULL;
	firstLink         = NULL;
	lastLink          = NULL;
	topLink           = NULL;
	selectedLink      = NULL;
	curNumStrings     = 0;
	justificationRule = Font::JR_left;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::load(Iff *iff)
{
	int   i;
	int   r, g, b;
	char  buffer[128];

	if (iff)
	{
		MenuElement::load(iff);

		// grab the justification rule
		if (iff->seekChunk(TAG_JSTY, Iff::FlagOptional))
		{
			int temp=iff->read_u_int32();
			switch (temp)
			{
				case 1:
					justificationRule=Font::JR_left;
					break;
				case 2:
					justificationRule=Font::JR_right;
					break;
				case 3:
					justificationRule=Font::JR_centerX;
					break;
				default:
					Fatal("MenuStaticListBox::load bad justify type (%d)", temp);
			}
			iff->exitChunk(); // 'JSTY'
		}

		if (iff->seekChunk(TAG_LBF, Iff::FlagOptional))
		{
			r = iff->read_int32();
			g = iff->read_int32();
			b = iff->read_int32();
			iff->read_string(buffer, sizeof(buffer));

			font = NEW Font(buffer,r,g,b);
			iff->exitChunk();
		}
		else
		{
			if (iff->seekChunk(TAG_LBFS, Iff::FlagOptional))
			{
				iff->read_string(buffer, sizeof(buffer));
				font = findSharedFont(buffer);
				DEBUG_FATAL(!font, ("MENU_LIST_BOX cannot find shared font '%s'", buffer));
				iff->exitChunk();
			}
			else
				font = NEW Font("font\\terminal.iff", 255, 255, 255);  // default to terminal in case none is specified
		}

		// load in the stringSelectEvent number
		numStringEvents = iff->blockCount(Iff::FlagChunk, TAG_LBSE);
		if (numStringEvents > 0)
		{
			stringSelectEvent = NEW int[numStringEvents];
			for (i = 0; iff->seekChunk(TAG_LBSE, Iff::FlagOptional); i++)
			{
				stringSelectEvent[i] = iff->read_int32();
				iff->exitChunk(); // 'LBSE'
			}
		}

		// do we want to override the font height?
		textHeight = font->pixelHeight();
		if (iff->seekChunk(TAG_LBTH, Iff::FlagOptional))
		{
			if ((textHeight = iff->read_int32()) <= 0)
				Fatal("MenuStaticListBox::load textHeight MUST be > 0 (%d was specified)", textHeight);
			iff->exitChunk(); // 'LBTH'
		}

		// specify a highlight font? (default to original font)
		fontHighlight = font;
		if (iff->seekChunk(TAG_LBHF, Iff::FlagOptional))
		{
			r = iff->read_int32();
			g = iff->read_int32();
			b = iff->read_int32();
			iff->read_string(buffer, sizeof(buffer));
			fontHighlight = NEW Font(buffer,r,g,b);
			iff->exitChunk();

		}
		else
		{
			if (iff->seekChunk(TAG_LBHS, Iff::FlagOptional))
			{
				iff->read_string(buffer, sizeof(buffer));
				fontHighlight = findSharedFont(buffer);
				DEBUG_FATAL(!font, ("MENU_LIST_BOX cannot find shared font '%s'", buffer));
				iff->exitChunk();
			}
		}

		// record off how many lines the user can see
		visibleLinks = (menuRect.y1 - menuRect.y0 + 1) / textHeight;

		topLink = firstLink;
		// initialize lastLink
		for (lastLink = firstLink; lastLink && lastLink->next; lastLink = lastLink->next)
			;
	}
}

// ----------------------------------------------------------------------
// hover event messages occur only once for each 'start' and 'stop'
IoResult MenuStaticListBox::event(IoEventStruct *event)
{
	int s;

	if (menuRect.isInInclusive(event->mouseX, event->mouseY))
	{
		// string selection?
		if ((stringSelectEvent) && event->type == IoEventMousePress && event->button == 0)
		{
			// compute the string offset that the user clicked on
			s = (event->mouseY - menuRect.y0 - 2) / textHeight;
			for (selectedLink = topLink; selectedLink && (--s >= 0); selectedLink = selectedLink->next)
				;

			for (s = 0; s < numStringEvents; s++)
				menuEvent(stringSelectEvent[s], this);
		}

		return IoResultBlock;
	}

	return MenuElement::event(event);
}

// ----------------------------------------------------------------------

void MenuStaticListBox::menuDraw(int x, int y) const
{
	Link *link;
	int  i;
	Rect r;
	Font *f;

	r.x0 = x + menuRect.x0;
	r.y0 = y + menuRect.y0;
	r.x1 = r.x0 + (menuRect.x1 - menuRect.x0 + 1);
	r.y1 = r.y0 + textHeight;
	for (link = topLink, i = 0; link && i < visibleLinks; link = link->next, i++)
	{
		if (link->string)
		{
			f = font;
			if (selectedLink == link)
				f = fontHighlight;
			else
				if (link->overridingFont)
					f = link->overridingFont;
			f->printString(*link->string, r.x0,r.y0,r.x1,justificationRule);
			r.y0 += textHeight;
			r.y1 += textHeight;
		}
	}

	MenuElement::menuDraw(x, y);
}

// ----------------------------------------------------------------------

void MenuStaticListBox::addLinkToEnd(Link *link)
{
	// if this is a dynamic linked list then we must ensure that it gets setup properly
	if (!firstLink)
	{
		firstLink = link;
		topLink   = firstLink;
	}
	else
		lastLink->next = link; // tack on the link to the end of the list

	link->prev      = lastLink;
	lastLink        = link;
	lastLink->next  = NULL;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::realAddString(const String &string, Font *overrideFont)
{
	int        pos;
	Link       *link;
	Font *f;

	f = font;
	if (overrideFont)
		f = overrideFont;

	link = NEW Link;
	addLinkToEnd(link);

	link->string = NEW String(string.getPointer());
	link->overridingFont = overrideFont;
	curNumStrings++;

	// do we need to scroll down one after this because we just inserted a string at the end of the list?
	for (pos = 0, link = topLink; link; link = link->prev, pos++)
		;
	if (pos == visibleLinks || ((curNumStrings - pos + 1) > visibleLinks))
		scrollDown();
}

// ----------------------------------------------------------------------

void MenuStaticListBox::addString(const String &string, Font *overrideFont)
{
	realAddString(string, overrideFont);
}

// ----------------------------------------------------------------------

void MenuStaticListBox::addString(int intToAdd, Font *overrideFont)
{
	char str[32]; // I think this is enough room

	realAddString(String(itoa(intToAdd, str, 10)), overrideFont);
}

// ----------------------------------------------------------------------

void MenuStaticListBox::scrollUp(void)
{
	topLink = (topLink && topLink->prev) ? topLink->prev : topLink;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::scrollDown(void)
{
	int  linksToTopLink;
	Link *link;

	// count the number of links left in the list
	for (link = topLink, linksToTopLink = 0; link; link = link->prev, linksToTopLink++)
		;

	if ((curNumStrings - linksToTopLink + 1) > visibleLinks)
		topLink = topLink->next;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::scrollAbsolute(real percentage)
{
	int topEntry;

	if (percentage >= CONST_REAL(1))
		percentage = 0.999f;

	topEntry = 0;
	if (curNumStrings >= visibleLinks)
		topEntry = static_cast<int>(floor(percentage * (curNumStrings - visibleLinks + 1)));

	for (topLink = firstLink; topEntry--; topLink = topLink->next)
		;
}

// ----------------------------------------------------------------------

real MenuStaticListBox::getPercentage(void) const
{
	real percentage = CONST_REAL(0);
	int  position;
	Link *link;

	if (curNumStrings >= visibleLinks)
	{
		for (link = firstLink, position = 0; link != topLink && link != lastLink; link = link->next, position++)
			;

		percentage =  static_cast<real>(position);
		percentage /= static_cast<real>(curNumStrings - visibleLinks);
	}

	return percentage;
}

// ----------------------------------------------------------------------

bool MenuStaticListBox::stringExists(const String &string) const
{
	bool found = false;
	Link *link;

	for (link = firstLink; link && !found; link = link->next)
		if (link->string && string==*link->string)
			found = true;

	return found;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::changeString(const String &oldString, const String &newString, Font *overrideFont)
{
	Link *link;

	for (link = firstLink; link; link = link->next)
		if (link->string && oldString==*link->string)
		{
			*link->string=newString;
			if (overrideFont)
				link->overridingFont = overrideFont;
		}
}

// ----------------------------------------------------------------------

void MenuStaticListBox::changeCurrentSelectedString(const String &newString, Font *overrideFont)
{
	if (selectedLink)
	{
		*selectedLink->string=newString;	
		if (overrideFont)
			selectedLink->overridingFont = overrideFont;
	}
}

// ----------------------------------------------------------------------

void MenuStaticListBox::removeAllStrings(void)
{
	while (firstLink)
	{
		Link *link;

		link=firstLink;
		firstLink=firstLink->next;

		delete link;
	}

	lastLink = NULL;
	selectedLink = NULL;
	topLink = NULL;
	curNumStrings = 0;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::removeString(const String &string)
{
	int  pos;
	Link *link;
	Link *delLink;
	int  topOffset;

	// search the list for the following:
	// 1) the topLink's offset into the list
	// 2) the actual link to delete (if any)
	for (link = firstLink, pos = 0, topOffset = 0, delLink = NULL; link && link->string; link = link->next, pos++)
	{
		if (link == topLink)
			topOffset = pos;

		// take the first occurrance of this string
		if (!delLink && link->string && string == *link->string)
			delLink = link;
	}

	// did we find it?
	if (delLink)
	{
		// ensure proper "top offset" positioning when removing a string
		if (topLink == delLink)
			topLink = topLink->prev ? topLink->prev : topLink->next;
		else
			if ((curNumStrings - topOffset) < visibleLinks && topLink->prev)
				topLink = topLink->prev;

		// extract the link out of the list
		if (delLink->prev)
			delLink->prev->next = delLink->next;
		else
			firstLink = delLink->next;
		if (delLink->next)
			delLink->next->prev = delLink->prev;
		else
			lastLink = delLink->prev;

		// if this link is the current selected link then unselect it
		if (selectedLink == delLink)
			selectedLink = NULL;

		delete delLink;
		curNumStrings--;
	}
}

// ----------------------------------------------------------------------

const String &MenuStaticListBox::getString(int absolutePosition) const
{
	Link         *link;
	int          i;

	// switched ASSERT to DEBUG_FATAL : should this be aP >= cNS? -ALS
	DEBUG_FATAL(absolutePosition < 0 || absolutePosition > curNumStrings,("MenuStaticListBox::getString - bogus absolute position %d.",absolutePosition));

	// the "absPos - 1" guarantees that when the link IS found that link = link->next happens on the node before
	for (i = 0, link = firstLink; link && i < absolutePosition; link = link->next, i++)
		;

	return *link->string;
}

// ----------------------------------------------------------------------
// return values: -1 if the string is not found, otherwise 
//                0 - "last string" if the string was found
int MenuStaticListBox::index(const String &string) const
{
	int  i;
	Link *link;
	int  result = -1;

	for (link = firstLink, i = 0, result = -1; link && (result == -1); link = link->next, i++)
		if (link->string && string==*link->string)
			result = i;

	return result;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::resetCurrentSelectedString(void)
{
	selectedLink = NULL;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::setCurrentSelectedString(const String &string)
{
	Link *link;

	for (selectedLink = NULL, link = firstLink; !selectedLink && link; link = link->next)
		if (link->string && string==*link->string)
			selectedLink = link;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::setCurrentSelectedString(int absoluteIndex)
{
	Link *link = NULL;
	int  i;

	if (absoluteIndex >= 0 && absoluteIndex <= curNumStrings)
		for (i = 0, link = firstLink; i < absoluteIndex && link; i++, link = link->next)
			;

	selectedLink = link;
}

// ----------------------------------------------------------------------

const String *MenuStaticListBox::getCurrentSelectedString(void) const
{
	return selectedLink ? selectedLink->string : NULL;
}

// ----------------------------------------------------------------------

int MenuStaticListBox::getCurrentSelectedStringOffset(void) const
{
	Link *link;
	bool found(false);
	int  i;

	i = -1;

	if (selectedLink != NULL)
		for (i = 0, link = firstLink; link; i++, link = link->next)
			if (selectedLink == link)
			{
				found = true;
				break;
			}

	if (found)
		return i;
	return -1;
}

// ----------------------------------------------------------------------

void MenuStaticListBox::setFont(Font *newFont)
{
	if (font == fontHighlight)
		font = fontHighlight = newFont;
	else
		font = newFont;
}

// ----------------------------------------------------------------------

const Font *MenuStaticListBox::getFont(void) const
{
	return font;
}

// ----------------------------------------------------------------------

